# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-09-23 15:14
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0002_add_batch_level_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='batch',
            name='custom_permissions',
            field=models.BooleanField(default=False),
        ),
    ]



from collections import defaultdict
from datetime import datetime, timedelta
from functools import wraps
import logging
import urllib

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db import connection, transaction
from django.db.utils import OperationalError
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.dateparse import parse_date
from django.utils.datastructures import MultiValueDictKeyError

from .models import Task, TaskAssignment, Batch, Project

User = get_user_model()

logger = logging.getLogger(__name__)


def handle_db_lock(func):
    """Decorator that catches database lock errors from sqlite"""
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        try:
            return func(request, *args, **kwargs)
        except OperationalError as ex:
            # sqlite3 cannot handle concurrent transactions.
            # This should be very rare with just a few users.
            # If it happens often, switch to mysql or postgres.
            if str(ex) == 'database is locked':
                messages.error(request, u'The database is busy. Please try again.')
                return redirect(index)
            raise ex
    return wrapper


def index(request):
    """
    Security behavior:
    - Anyone can access the page, but the page only shows the user
      information they have access to.
    """
    abandoned_assignments = []
    if request.user.is_authenticated:
        for ha in TaskAssignment.objects.filter(assigned_to=request.user)\
                                        .filter(completed=False)\
                                        .filter(task__batch__active=True)\
                                        .filter(task__batch__project__active=True):
            abandoned_assignments.append({
                'task': ha.task,
                'task_assignment_id': ha.id
            })

    batch_list = Batch.access_permitted_for(request.user)
    batch_query = Batch.objects.filter(id__in=[b.id for b in batch_list])

    available_task_counts = Batch.available_task_counts_for(batch_query, request.user)

    batch_rows = []
    for batch in batch_query.values('created_at', 'id', 'name', 'project__name'):
        total_tasks_available = available_task_counts[batch['id']]

        if total_tasks_available > 0:
            batch_rows.append({
                'project_name': batch['project__name'],
                'batch_name': batch['name'],
                'batch_published': batch['created_at'],
                'assignments_available': total_tasks_available,
                'preview_next_task_url': reverse('preview_next_task',
                                                 kwargs={'batch_id': batch['id']}),
                'accept_next_task_url': reverse('accept_next_task',
                                                kwargs={'batch_id': batch['id']})
            })
    return render(request, 'turkle/index.html', {
        'abandoned_assignments': abandoned_assignments,
        'batch_rows': batch_rows
    })


@handle_db_lock
def accept_task(request, batch_id, task_id):
    """
    Accept task from preview

    Security behavior:
    - If the user does not have permission to access the Batch+Task, they
      are redirected to the index page with an error message.
    """
    try:
        batch = Batch.objects.get(id=batch_id)
    except ObjectDoesNotExist:
        messages.error(request, u'Cannot find Task Batch with ID {}'.format(batch_id))
        return redirect(index)
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, u'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)

    try:
        with transaction.atomic():
            # Lock access to the specified Task
            # len() forces execution of query to create lock
            len(Task.objects.filter(id=task_id).select_for_update())

            # Will throw ObjectDoesNotExist exception if Task no longer available
            batch.available_tasks_for(request.user).get(id=task_id)

            ha = TaskAssignment()
            if request.user.is_authenticated:
                ha.assigned_to = request.user
            else:
                ha.assigned_to = None
            ha.task = task
            ha.save()
            if request.user.is_authenticated:
                logger.info('User(%i) accepted Task(%i)', request.user.id, task.id)
            else:
                logger.info('Anonymous user accepted Task(%i)', task.id)
    except ObjectDoesNotExist:
        messages.error(request, u'The Task with ID {} is no longer available'.format(task_id))
        return redirect(index)

    return redirect(task_assignment, task.id, ha.id)


@handle_db_lock
def accept_next_task(request, batch_id):
    """
    Accept task from index or auto accept next task

    Security behavior:
    - If the user does not have permission to access the Batch+Task, they
      are redirected to the index page with an error message.
    """
    try:
        with transaction.atomic():
            batch = Batch.objects.get(id=batch_id)

            # Lock access to all Tasks available to current user in the batch
            # Force evaluation of the query set with len()
            # But postgres does not support select_for_update() with outer join
            if connection.vendor != "postgresql":
                len(batch.available_task_ids_for(request.user).select_for_update())
            else:
                # Locks a few more tasks for multiple annotation batches than above
                len(batch.unfinished_tasks().select_for_update())

            task_id = _skip_aware_next_available_task_id(request, batch)

            if task_id:
                ha = TaskAssignment()
                if request.user.is_authenticated:
                    ha.assigned_to = request.user
                else:
                    ha.assigned_to = None
                ha.task_id = task_id
                ha.save()
                if request.user.is_authenticated:
                    logger.info('User(%i) accepted Task(%i)', request.user.id, task_id)
                else:
                    logger.info('Anonymous user accepted Task(%i)', task_id)
    except ObjectDoesNotExist:
        messages.error(request, u'Cannot find Task Batch with ID {}'.format(batch_id))
        return redirect(index)

    if task_id:
        return redirect(task_assignment, task_id, ha.id)
    else:
        messages.error(request, u'No more Tasks available for Batch {}'.format(batch.name))
        return redirect(index)


def help_page(request):
    return render(request, 'turkle/help.html')


def task_assignment(request, task_id, task_assignment_id):
    """
    Task view and submission (task content in iframe below)

    Security behavior:
    - If the user does not have permission to access the Task Assignment, they
      are redirected to the index page with an error message.
    """
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)
    try:
        task_assignment = TaskAssignment.objects.get(id=task_assignment_id)
    except ObjectDoesNotExist:
        messages.error(request,
                       'Cannot find Task Assignment with ID {}'.format(task_assignment_id))
        return redirect(index)

    if request.user.is_authenticated:
        if request.user != task_assignment.assigned_to:
            messages.error(
                request,
                'You do not have permission to work on the Task Assignment with ID {}'.
                format(task_assignment.id))
            return redirect(index)
    else:
        if task_assignment.assigned_to is not None:
            messages.error(
                request,
                'You do not have permission to work on the Task Assignment with ID {}'.
                format(task_assignment.id))
            return redirect(index)

    auto_accept_status = request.session.get('auto_accept_status', False)

    if request.method == 'GET':
        http_get_params = "?assignmentId={}&hitId={}&workerId={}&urlSubmitTo={}".format(
            task_assignment.id,
            task.id,
            request.user.id,
            urllib.parse.quote(
                reverse('task_assignment', kwargs={
                    'task_id': task.id, 'task_assignment_id': task_assignment.id}),
                safe=''))
        return render(
            request,
            'turkle/task_assignment.html',
            {
                'auto_accept_status': auto_accept_status,
                'http_get_params': http_get_params,
                'task': task,
                'task_assignment': task_assignment,
            },
        )
    else:
        task_assignment.answers = dict(request.POST.items())
        task_assignment.completed = True
        task_assignment.save()
        if request.user.is_authenticated:
            logger.info('User(%i) submitted Task(%i)', request.user.id, task.id)
        else:
            logger.info('Anonymous user submitted Task(%i)', task.id)

        if request.session.get('auto_accept_status'):
            return redirect(accept_next_task, task.batch.id)
        else:
            return redirect(index)


def task_assignment_iframe(request, task_id, task_assignment_id):
    """
    Task form served in an iframe

    Security behavior:
    - If the user does not have permission to access the Task Assignment, they
      are redirected to the index page with an error messge.
    """
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)
    try:
        task_assignment = TaskAssignment.objects.get(id=task_assignment_id)
    except ObjectDoesNotExist:
        messages.error(request,
                       'Cannot find Task Assignment with ID {}'.format(task_assignment_id))
        return redirect(index)

    if request.user.is_authenticated:
        if request.user != task_assignment.assigned_to:
            messages.error(
                request,
                'You do not have permission to work on the Task Assignment with ID {}'.
                format(task_assignment.id))
            return redirect(index)

    return render(
        request,
        'turkle/task_assignment_iframe.html',
        {
            'task': task,
            'task_assignment': task_assignment,
        },
    )


def preview(request, task_id):
    """
    Security behavior:
    - If the user does not have permission to access the Task, they
      are redirected to the index page with an error message.
    """
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)

    if not task.batch.project.available_for(request.user):
        messages.error(request, 'You do not have permission to view this Task')
        return redirect(index)

    http_get_params = "?assignmentId=ASSIGNMENT_ID_NOT_AVAILABLE&hitId={}".format(
        task.id)
    return render(request, 'turkle/preview.html', {
        'http_get_params': http_get_params,
        'task': task
    })


def preview_iframe(request, task_id):
    """
    Security behavior:
    - If the user does not have permission to access the Task, they
      are redirected to the index page with an error message.
    """
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)

    if not task.batch.project.available_for(request.user):
        messages.error(request, 'You do not have permission to view this Task')
        return redirect(index)

    return render(request, 'turkle/preview_iframe.html', {'task': task})


def preview_next_task(request, batch_id):
    """
    Security behavior:
    - If the user does not have permission to access the Batch, they
      are redirected to the index page with an error message.
    """
    try:
        batch = Batch.objects.get(id=batch_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task Batch with ID {}'.format(batch_id))
        return redirect(index)

    task_id = _skip_aware_next_available_task_id(request, batch)

    if task_id:
        return redirect(preview, task_id)
    else:
        messages.error(request,
                       'No more Tasks are available for Batch "{}"'.format(batch.name))
        return redirect(index)


def return_task_assignment(request, task_id, task_assignment_id):
    """
    Security behavior:
    - If the user does not have permission to return the Assignment, they
      are redirected to the index page with an error message.
    """
    redirect_due_to_error = _delete_task_assignment(request, task_id, task_assignment_id)
    if redirect_due_to_error:
        return redirect_due_to_error
    if request.user.is_authenticated:
        logger.info('User(%i) returned Task(%i)', request.user.id, int(task_id))
    else:
        logger.info('Anonymous user returned Task(%i)', int(task_id))
    return redirect(index)


def skip_and_accept_next_task(request, batch_id, task_id, task_assignment_id):
    """
    Security behavior:
    - If the user does not have permission to return the Assignment, they
      are redirected to the index page with an error message.
    """
    redirect_due_to_error = _delete_task_assignment(request, task_id, task_assignment_id)
    if redirect_due_to_error:
        return redirect_due_to_error

    _add_task_id_to_skip_session(request.session, batch_id, task_id)
    if request.user.is_authenticated:
        logger.info('User(%i) skipped Task(%i)', request.user.id, int(task_id))
    else:
        logger.info('Anonymous user skipped Task(%i)', int(task_id))
    return redirect(accept_next_task, batch_id)


def skip_task(request, batch_id, task_id):
    """
    Skip to next task when previewing a task

    Security behavior:
    - This view updates a session variable that controls the order
      that Tasks are presented to a user.  Users cannot modify other
      users session variables.
    """
    _add_task_id_to_skip_session(request.session, batch_id, task_id)
    return redirect(preview_next_task, batch_id)


def stats_for_self(request):
    if not request.user.is_authenticated:
        messages.error(
            request,
            u'You must be logged in to view the user statistics page')
        return redirect(index)

    return stats_for_user(request, request.user.id)


def parse_date_with_timezone(value):
    """wrapper for Django's parse_date() that uses timezone when appropriate"""
    date = parse_date(value)
    dt = datetime(date.year, date.month, date.day)
    if settings.USE_TZ:
        dt = timezone.make_aware(dt)
    return dt


def stats_for_user(request, user_id):
    def format_seconds(s):
        """Converts seconds to string"""
        return '%dh %dm' % (s//3600, (s//60) % 60)

    try:
        user = User.objects.get(id=user_id)
    except ObjectDoesNotExist:
        messages.error(request, u'Cannot find User with ID {}'.format(user_id))
        return redirect(index)

    if request.user.id != user_id and not request.user.is_staff:
        messages.error(request, u"You cannot view another User's statistics unless you are Staff")
        return redirect(index)

    try:
        start_date = parse_date_with_timezone(request.GET['start_date'])
    except MultiValueDictKeyError:
        start_date = None
    try:
        end_date = parse_date_with_timezone(request.GET['end_date'])
    except MultiValueDictKeyError:
        end_date = None

    tas = TaskAssignment.objects.filter(completed=True).filter(assigned_to=user)
    if start_date:
        tas = tas.filter(updated_at__gte=start_date)
    if end_date:
        # adds a day to include assignments completed on the selected end date
        tas = tas.filter(updated_at__lte=end_date + timedelta(days=1))

    projects = Project.objects.filter(batch__task__taskassignment__assigned_to=user).\
        distinct()
    batches = Batch.objects.filter(task__taskassignment__assigned_to=user).distinct()

    elapsed_seconds_overall = 0
    project_stats = []
    for project in projects:
        project_batches = [b for b in batches if b.project_id == project.id]

        batch_stats = []
        elapsed_seconds_project = 0
        total_completed_project = 0
        for batch in project_batches:
            batch_tas = tas.filter(task__batch=batch)
            total_completed_batch = batch_tas.count()
            total_completed_project += total_completed_batch
            elapsed_seconds_batch = sum([ta.work_time_in_seconds() for ta in batch_tas])
            elapsed_seconds_project += elapsed_seconds_batch
            elapsed_seconds_overall += elapsed_seconds_batch
            if total_completed_batch > 0:
                batch_stats.append({
                    'batch_name': batch.name,
                    'elapsed_time_batch': format_seconds(elapsed_seconds_batch),
                    'total_completed_batch': total_completed_batch,
                })
        if total_completed_project > 0:
            project_stats.append({
                'project_name': project.name,
                'batch_stats': batch_stats,
                'elapsed_time_project': format_seconds(elapsed_seconds_project),
                'total_completed_project': total_completed_project,
            })

    if start_date:
        start_date = start_date.strftime('%Y-%m-%d')
    if end_date:
        end_date = end_date.strftime('%Y-%m-%d')

    name = user.get_full_name()
    if not name:
        name = user.get_username()

    return render(
        request,
        'turkle/stats.html',
        {
            'project_stats': project_stats,
            'end_date': end_date,
            'start_date': start_date,
            'total_completed': tas.count(),
            'total_elapsed_time': format_seconds(elapsed_seconds_overall),
            'full_name': name,
            'user_id': user.id
        }
    )


def update_auto_accept(request):
    """
    Security behavior:
    - This view updates a session variable that controls whether or
      not Task Assignments are auto-accepted.  Users cannot modify other
      users session variables.
    """
    accept_status = (request.POST['auto_accept'] == 'true')
    request.session['auto_accept_status'] = accept_status
    return JsonResponse({})


def user_activity_json(request, user_id):
    if request.user.id != user_id and not request.user.is_staff:
        return JsonResponse({})

    try:
        user = User.objects.get(id=user_id)
    except ObjectDoesNotExist:
        return JsonResponse({})

    # Create dictionary mapping timestamp (in seconds) to number of TaskAssignments
    # completed at that timestamp
    completed_at = TaskAssignment.objects.\
        filter(completed=True).\
        filter(assigned_to=user).\
        values_list('updated_at', flat=True)
    timestamp_counts = defaultdict(int)
    for ca in completed_at:
        timestamp_counts[int(ca.timestamp())] += 1

    return JsonResponse(timestamp_counts)


def _add_task_id_to_skip_session(session, batch_id, task_id):
    """Add Task ID to session variable tracking Tasks the user has skipped
    """
    # The Django session store converts dictionary keys from ints to strings
    batch_id = str(batch_id)
    task_id = str(task_id)

    if 'skipped_tasks_in_batch' not in session:
        session['skipped_tasks_in_batch'] = {}
    if batch_id not in session['skipped_tasks_in_batch']:
        session['skipped_tasks_in_batch'][batch_id] = []
        session.modified = True
    if task_id not in session['skipped_tasks_in_batch'][batch_id]:
        session['skipped_tasks_in_batch'][batch_id].append(task_id)
        session.modified = True


@handle_db_lock
def _delete_task_assignment(request, task_id, task_assignment_id):
    """Delete a TaskAssignment, if possible

    Returns:
        - None if the TaskAssignment can be deleted, *OR*
        - An HTTPResponse object created by redirect() if there was an error

    Usage:
        redirect_due_to_error = _delete_task_assignment(request, task_id, task_assignment_id)
        if redirect_due_to_error:
            return redirect_due_to_error
    """
    try:
        task = Task.objects.get(id=task_id)
    except ObjectDoesNotExist:
        messages.error(request, 'Cannot find Task with ID {}'.format(task_id))
        return redirect(index)
    try:
        task_assignment = TaskAssignment.objects.get(id=task_assignment_id)
    except ObjectDoesNotExist:
        messages.error(request,
                       'Cannot find Task Assignment with ID {}'.format(task_assignment_id))
        return redirect(index)

    if task_assignment.completed:
        messages.error(request, u"The Task can't be returned because it has been completed")
        return redirect(index)
    if request.user.is_authenticated:
        if task_assignment.assigned_to != request.user:
            messages.error(request, 'The Task you are trying to return belongs to another user')
            return redirect(index)
    else:
        if task_assignment.assigned_to is not None:
            messages.error(request, 'The Task you are trying to return belongs to another user')
            return redirect(index)
        if task.batch.project.login_required:
            messages.error(request, 'You do not have permission to access this Task')
            return redirect(index)

    task_assignment.delete()


def _skip_aware_next_available_task_id(request, batch):
    """Get next available Task for user, taking into account previously skipped Tasks

    This function will first look for an available Task that the user
    has not previously skipped.  If the only available Tasks are Tasks
    that the user has skipped, this function will return the first
    such Task.

    Returns:
        Task ID (int), or None if no more Tasks are available
    """
    def _get_skipped_task_ids_for_batch(session, batch_id):
        batch_id = str(batch_id)
        if 'skipped_tasks_in_batch' in session and \
           batch_id in session['skipped_tasks_in_batch']:
            return session['skipped_tasks_in_batch'][batch_id]
        else:
            return None

    available_task_ids = batch.available_task_ids_for(request.user)
    skipped_ids = _get_skipped_task_ids_for_batch(request.session, batch.id)

    if skipped_ids:
        task_id = available_task_ids.exclude(id__in=skipped_ids).first()
        if not task_id:
            task_id = available_task_ids.filter(id__in=skipped_ids).first()
            if task_id:
                messages.info(request, 'Only previously skipped Tasks are available')

                # Once all remaining Tasks have been marked as skipped, we clear
                # their skipped status.  If we don't take this step, then a Task
                # cannot be skipped a second time.
                request.session['skipped_tasks_in_batch'][str(batch.id)] = []
                request.session.modified = True
    else:
        task_id = available_task_ids.first()

    return task_id



# Generated by Django 2.2.13 on 2020-07-02 23:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0006_copy_can_work_on_permission'),
    ]

    operations = [
        migrations.AlterField(
            model_name='batch',
            name='created_by',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='created_batches', to=settings.AUTH_USER_MODEL, verbose_name='creator'),
        ),
        migrations.AlterField(
            model_name='project',
            name='created_by',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='created_projects', to=settings.AUTH_USER_MODEL, verbose_name='creator'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-09-17 15:37
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='batch',
            options={'permissions': (('can_work_on_batch', 'Can work on Tasks for this Batch'),), 'verbose_name': 'Batch', 'verbose_name_plural': 'Batches'},
        ),
    ]



import csv
import ctypes
from datetime import timedelta
import logging
import os.path
import re
import statistics
import sys

from bs4 import BeautifulSoup
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count, IntegerField, Max, Q, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce
from django.utils import timezone
from guardian.core import ObjectPermissionChecker
from guardian.models import GroupObjectPermission
from guardian.shortcuts import assign_perm, get_group_perms, get_groups_with_perms
from jsonfield import JSONField

from .utils import get_turkle_template_limit

User = get_user_model()

logger = logging.getLogger(__name__)

C_LONG_NUM_BITS = 8 * ctypes.sizeof(ctypes.c_long)
C_LONG_MAX = 2 ** (C_LONG_NUM_BITS-1) - 1

# Increase default field size limit (default 131072 characters).
# Note that field_size_limit converts its argument to a C long,
# at least in Anaconda 3 on Windows 10.
csv.field_size_limit(min(C_LONG_MAX, sys.maxsize))


class ActiveUserManager(models.Manager):
    """Query users by activity on assignments"""
    def get_queryset(self, **kwargs):
        # adds annotations for number of assignments and most recent assignment time
        n_days = int(kwargs.get('n_days', 7))
        time_cutoff = timezone.now() - timedelta(days=n_days)
        active_users = super().get_queryset(). \
            filter(Q(taskassignment__updated_at__gt=time_cutoff) &
                   Q(taskassignment__completed=True)). \
            distinct(). \
            annotate(
                total_assignments=Count('taskassignment',
                                        filter=(Q(taskassignment__updated_at__gt=time_cutoff) &
                                                Q(taskassignment__completed=True)))). \
            annotate(last_finished_time=Max('taskassignment__updated_at',
                                            filter=Q(taskassignment__completed=True)))
        return active_users


class ActiveUser(User):
    """Proxy object for User that provides fields related to assignments.

    The values for the additional fields come from annotations set by the manager.
    """
    objects = ActiveUserManager()

    class Meta:
        proxy = True
        ordering = ['first_name']

    def name(self):
        return f"{self.first_name} {self.last_name}"
    name.admin_order_field = 'first_name'

    def completed_assignments(self):
        return self.total_assignments
    completed_assignments.admin_order_field = 'total_assignments'

    def most_recent(self):
        return self.last_finished_time
    most_recent.admin_order_field = 'last_finished_time'


class TaskAssignmentStatistics(object):
    """Mixin class for Batch/Project that computes TaskAssignment statistics

    Assumes that the inheriting class has a finished_task_assignments()
    method that returns a QuerySet of TaskAssignments.
    """
    def mean_work_time_in_seconds(self):
        """
        Returns:
            Float for mean work time (in seconds) for completed Tasks in this Batch
        """
        finished_assignments = self.finished_task_assignments()
        if finished_assignments.count() > 0:
            return statistics.mean(
                [ta.work_time_in_seconds() for ta in self.finished_task_assignments()])
        else:
            return 0

    def median_work_time_in_seconds(self):
        """
        Returns:
            Integer for median work time (in seconds) for completed Tasks in this Batch
        """
        if self.finished_task_assignments().count() > 0:
            # np.median returns float but we convert back to int computed by work_time_in_seconds()
            return int(statistics.median(
                [ta.work_time_in_seconds() for ta in self.finished_task_assignments()]
            ))
        else:
            return 0

    def total_work_time_in_seconds(self):
        """
        Returns:
            Integer sum of work_time_in_seconds() for all completed
            TaskAssignments in this Batch
        """
        return sum([ta.work_time_in_seconds() for ta in self.finished_task_assignments()])


class Task(models.Model):
    """Human Intelligence Task
    """
    class Meta:
        verbose_name = "Task"

    batch = models.ForeignKey('Batch', on_delete=models.CASCADE)
    completed = models.BooleanField(default=False)
    input_csv_fields = JSONField()

    def __str__(self):
        return 'Task id:{}'.format(self.id)

    def populate_html_template(self):
        """Return HTML template for this Task's project, with populated template variables

        Returns:
            String containing the HTML template for the Project associated with
            this Task, with all template variables replaced with the template
            variable values stored in this Task's input_csv_fields.
        """
        result = self.batch.project.html_template
        for field in self.input_csv_fields.keys():
            result = result.replace(
                r'${' + field + r'}',
                self.input_csv_fields[field]
            )
        return result


class TaskAssignment(models.Model):
    """Task Assignment
    """
    class Meta:
        verbose_name = "Task Assignment"

    answers = JSONField(blank=True)
    assigned_to = models.ForeignKey(settings.AUTH_USER_MODEL, db_index=True, null=True,
                                    on_delete=models.CASCADE)
    completed = models.BooleanField(db_index=True, default=False)
    created_at = models.DateTimeField(auto_now_add=True)
    expires_at = models.DateTimeField(null=True)
    task = models.ForeignKey(Task, on_delete=models.CASCADE)
    updated_at = models.DateTimeField(auto_now=True, db_index=True)

    @classmethod
    def expire_all_abandoned(cls):
        result = cls.objects.\
            filter(completed=False).\
            filter(expires_at__lt=timezone.now()).\
            delete()
        logger.info('Expired %i task assignments', result[0])
        return result

    def save(self, *args, **kwargs):
        # set expires_at only when assignment is created
        if not self.id:
            self.expires_at = timezone.now() + \
                timedelta(hours=self.task.batch.allotted_assignment_time)

        if 'csrfmiddlewaretoken' in self.answers:
            del self.answers['csrfmiddlewaretoken']
        super().save(*args, **kwargs)

        # Mark Task as completed if all Assignments have been completed
        if self.task.taskassignment_set.filter(completed=True).count() >= \
           self.task.batch.assignments_per_task:
            self.task.completed = True
            self.task.save()

        # Mark Batch as completed if all Tasks have been completed
        self.task.batch.update_completed_status()

    def work_time_in_seconds(self):
        """Return number of seconds elapsed between Task assignment and submission

        We compute the time elapsed in Python instead of in SQL
        because "there are no native date/time fields in SQLite and
        Django currently emulates these features using a text field,"
        per the Django Docs:
          https://docs.djangoproject.com/en/3.1/ref/models/querysets/#aggregation-functions

        Returns:
            Integer for seconds elapsed between Task assignment and submission

        Raises:
            ValueError if TaskAssignment is not completed
        """
        if self.completed:
            return int((self.updated_at - self.created_at).total_seconds())
        else:
            raise ValueError(
                'Cannot compute work_time_in_seconds for incomplete TaskAssignment %d' %
                self.id)


class Batch(TaskAssignmentStatistics, models.Model):
    class Meta:
        permissions = (
            ('can_work_on_batch', 'Can work on Tasks for this Batch'),
        )
        verbose_name = "Batch"
        verbose_name_plural = "Batches"

    active = models.BooleanField(db_index=True, default=True)
    allotted_assignment_time = models.IntegerField(default=24)
    assignments_per_task = models.IntegerField(default=1, verbose_name='Assignments per Task')
    completed = models.BooleanField(default=False)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                   related_name='created_batches', on_delete=models.CASCADE,
                                   verbose_name='creator')
    custom_permissions = models.BooleanField(default=False)
    filename = models.CharField(max_length=1024)
    login_required = models.BooleanField(db_index=True, default=True)
    name = models.CharField(max_length=1024)
    project = models.ForeignKey('Project', on_delete=models.CASCADE)
    published = models.BooleanField(db_index=True, default=True)

    @classmethod
    def access_permitted_for(cls, user):
        """Retrieve the active Batches that the user has permission to access

        Both the Batch and the Project associated the Batch must be active.

        Args:
            user (User):

        Returns:
            List of Batch objects this user can access
        """
        batches = cls.objects.filter(active=True).filter(published=True)\
            .filter(project__active=True)
        if not user.is_authenticated:
            batches = batches.filter(login_required=False)

        # Can implement our own prefetch_perms() in future for efficiency
        checker = TurklePermissionChecker(user)
        checker.prefetch_perms(batches)

        return [b for b in batches if checker.has_perm('can_work_on_batch', b)]

    @classmethod
    def available_task_counts_for(cls, batch_query, user):
        """Retrieve # of tasks available for user for the Batches in query

        Args:
            batch_query (QuerySet): A QuerySet that retrieves Batch objects
            user (User):

        Returns:
            Dict where keys are Batch IDs (int) and values are the total
            number of tasks in the batch available for the specified user.
        """
        available_task_counts = {}

        # Completed batches (in the original batch_query) have an available task count of 0
        for b_id in batch_query.filter(completed=True).values_list('id', flat=True):
            available_task_counts[b_id] = 0

        if not user.is_authenticated:
            # Batches an anonymous user does not have access to have an available task count of 0
            for b_id in batch_query.filter(login_required=True).values_list('id', flat=True):
                available_task_counts[b_id] = 0
            batch_query = batch_query.exclude(login_required=True)

        # Count number of tasks available for case where Batch.assignments_per_task == 1
        # For this case, the number of available tasks is the same for all users with
        # access to the batch.
        oneway_batch_query = batch_query.filter(assignments_per_task=1).filter(completed=False)
        unassigned_tasks = Task.objects.filter(completed=False).filter(taskassignment=None)

        # Django does not easily support aggregations (such as Count) using subqueries:
        #   https://code.djangoproject.com/ticket/28296
        # The Django documentation states that:
        #  "Aggregates may be used within a Subquery, but they require a specific
        #   combination of filter(), values(), and annotate() to get the subquery
        #   grouping correct."
        #     https://docs.djangoproject.com/en/3.1/ref/models/expressions/#using-aggregates-within-a-subquery-expression
        # The specific syntax we use here for count subqueries is adapted from:
        #   https://github.com/martsberger/django-sql-utils
        # where we are using the pattern:
        #   subquery = Subquery(Child.objects.filter(parent_id=OuterRef('id')).order_by()
        #                      .values('parent').annotate(count=Count('pk'))
        #                      .values('count'), output_field=IntegerField())
        #   Parent.objects.annotate(child_count=Coalesce(subquery, 0))
        task_count_subquery = Subquery(
            unassigned_tasks
            .filter(batch=OuterRef('pk'))
            .order_by().values('batch').annotate(count=Count('pk')).values('count'),
            output_field=IntegerField())

        oneway_batch_values = oneway_batch_query\
            .annotate(available_task_count=Coalesce(task_count_subquery, 0))\
            .values('available_task_count', 'id')
        for obv in oneway_batch_values:
            available_task_counts[obv['id']] = obv['available_task_count']

        multiway_batch_query = batch_query.filter(assignments_per_task__gt=1)\
                                          .filter(completed=False)
        if user.is_authenticated:
            # Count number of tasks available for case where Batch.assignments_per_task > 1
            ta_count_subquery = Subquery(
                TaskAssignment.objects
                .filter(task=OuterRef('pk'))
                .order_by().values('task').annotate(count=Count('pk')).values('count'),
                output_field=IntegerField())
            unassigned_tasks = Task.objects.filter(completed=False)\
                                           .annotate(ac=Coalesce(ta_count_subquery, 0))\
                                           .filter(ac__lt=OuterRef('assignments_per_task'))\
                                           .exclude(taskassignment__assigned_to=user)
            task_count_subquery = Subquery(
                unassigned_tasks
                .filter(batch=OuterRef('pk'))
                .order_by().values('batch').annotate(count=Count('pk')).values('count'),
                output_field=IntegerField())
            multiway_batch_values = multiway_batch_query\
                .annotate(available_task_count=Coalesce(task_count_subquery, 0))\
                .values('available_task_count', 'id')
            for mbv in multiway_batch_values:
                available_task_counts[mbv['id']] = mbv['available_task_count']
        else:
            # Only authenticated users should have access to multiple-assignment batches
            # If the database somehow contains multiple-assignment Batches that are
            # accessible to an anonymous user, we set the available task count to 0
            for mbv in multiway_batch_query.values('id'):
                available_task_counts[mbv['id']] = 0

        return available_task_counts

    def assignments_completed_by(self, user):
        """
        Returns:
            QuerySet of all TaskAssignments completed by specified user
            that are part of this Batch
        """
        return TaskAssignment.objects.\
            filter(completed=True).\
            filter(assigned_to_id=user.id).\
            filter(task__batch=self)

    def available_for(self, user):
        """
        Returns:
            Boolean indicating if this Batch is available for the user
        """
        if not self.active or \
           not self.project.active or \
           (not user.is_authenticated and self.login_required):
            return False
        elif self.custom_permissions:
            return user.has_perm('can_work_on_batch', self)
        else:
            return True

    def available_tasks_for(self, user):
        """Retrieve a list of all Tasks in this Batch available for the user.

        This list DOES NOT include Tasks in the Batch that have been assigned
        to the user but not yet completed.

        Args:
            user (User|AnonymousUser):

        Returns:
            QuerySet of available Task objects
        """
        if not self.available_for(user):
            return Task.objects.none()

        hs = self.task_set.filter(completed=False)

        if self.assignments_per_task > 1:
            # Exclude Tasks that have already been assigned to this user.
            if user.is_authenticated:
                # If the user is not authenticated, then user.id is None,
                # and the query below would exclude all uncompleted Tasks.
                hs = hs.exclude(taskassignment__assigned_to_id=user.id)

            # Only include Tasks when # of (possibly incomplete) assignments < assignments_per_task
            hs = hs.annotate(ac=models.Count('taskassignment')).\
                filter(ac__lt=self.assignments_per_task)
        elif self.assignments_per_task == 1:
            # Only returns Tasks that have not been assigned to anyone (including this user)
            hs = hs.filter(taskassignment=None)

        return hs

    def available_task_ids_for(self, user):
        return self.available_tasks_for(user).values_list('id', flat=True)

    def clean(self):
        if not self.login_required and self.assignments_per_task != 1:
            raise ValidationError('When login is not required to access a Batch, ' +
                                  'the number of Assignments per Task must be 1')

    def copy_project_permissions(self):
        """Copy 'permission' settings from associated Project to this Batch

        Copies:
        - `custom_permissions` flag
        - `login_required` flag
        - group-level access permissions
        """
        self.custom_permissions = self.project.custom_permissions
        self.login_required = self.project.login_required
        if self.custom_permissions:
            for group in get_groups_with_perms(self.project):
                if 'can_work_on' in get_group_perms(group, self.project):
                    assign_perm('can_work_on_batch', group, self)

    def csv_results_filename(self):
        """Returns filename for CSV results file for this Batch
        """
        batch_filename, extension = os.path.splitext(os.path.basename(self.filename))

        # We are following Mechanical Turk's naming conventions for results files
        return "{}-Batch_{}_results{}".format(batch_filename, self.id, extension)

    def create_tasks_from_csv(self, csv_fh):
        """
        Args:
            csv_fh (file-like object): File handle for CSV input

        Returns:
            Number of Tasks created from CSV file
        """
        header, data_rows = self._parse_csv(csv_fh)

        logger.info('Creating tasks for Batch(%i) %s', self.id, self.name)
        num_created_tasks = 0
        for row in data_rows:
            if not row:
                continue
            task = Task(
                batch=self,
                input_csv_fields=dict(zip(header, row)),
            )
            task.save()
            num_created_tasks += 1
        logger.info('Created %i tasks for Batch(%i) %s', num_created_tasks, self.id, self.name)

        return num_created_tasks

    def finished_tasks(self):
        """
        Returns:
            QuerySet of all Task objects associated with this Batch
            that have been completed.
        """
        return self.task_set.filter(completed=True).order_by('-id')

    def finished_task_assignments(self):
        """
        Returns:
            QuerySet of all Task Assignment objects associated with this Batch
            that have been completed.
        """
        return TaskAssignment.objects.filter(task__batch_id=self.id)\
                                     .filter(completed=True)

    def is_active(self):
        return self.active and self.published
    is_active.short_description = 'Active'
    is_active.boolean = True

    def next_available_task_for(self, user):
        """Returns next available Task for the user, or None if no Tasks available

        Args:
            user (User):

        Returns:
            Task|None
        """
        return self.available_tasks_for(user).first()

    def total_assignments_completed_by(self, user):
        """
        Returns:
            Integer of total number of TaskAssignments completed by
            specified user that are part of this Batch
        """
        return self.assignments_completed_by(user).count()

    def total_available_tasks_for(self, user):
        """Returns number of Tasks available for the user

        Args:
            user (User):

        Returns:
            Number of Tasks available for user
        """
        return self.available_tasks_for(user).count()

    def total_finished_tasks(self):
        return self.finished_tasks().count()
    total_finished_tasks.short_description = 'Total finished Tasks'

    def total_finished_task_assignments(self):
        return self.finished_task_assignments().count()
    total_finished_task_assignments.short_description = 'Total finished Task Assignments'

    def total_task_assignments(self):
        return self.assignments_per_task * self.total_tasks()

    def total_tasks(self):
        return self.task_set.count()
    total_tasks.short_description = 'Total Tasks'

    def total_users_that_completed_tasks(self):
        """
        Returns:
            Integer counting number of Users that have completed
            TaskAssignments that are part of this Batch
        """
        return self.users_that_completed_tasks().count()

    def to_csv(self, csv_fh, lineterminator='\r\n'):
        """Write CSV output to file handle for every Task in batch

        Args:
            csv_fh (file-like object): File handle for CSV output
        """
        fieldnames, rows = self._results_data(self.task_set.all())
        writer = csv.DictWriter(csv_fh, fieldnames, lineterminator=lineterminator,
                                quoting=csv.QUOTE_ALL)
        writer.writeheader()
        for row in rows:
            writer.writerow(row)

    def to_input_csv(self, csv_fh, lineterminator='\r\n'):
        """Write (reconstructed) CSV input to file handle for every Task in Batch

        PLEASE NOTE: The column order in the reconstructed CSV file
        may not match the column order in the original CSV file.
        """
        tasks = self.task_set.all()
        if not tasks.exists():
            return

        # Some rows may (theoretically) be missing fields
        fieldnames = set()
        for task in tasks:
            fieldnames.update(task.input_csv_fields.keys())

        writer = csv.DictWriter(csv_fh, list(fieldnames), lineterminator=lineterminator,
                                quoting=csv.QUOTE_ALL)
        writer.writeheader()
        for task in tasks:
            writer.writerow(task.input_csv_fields)

    def unfinished_task_assignments(self):
        """
        Returns:
            QuerySet of all Task Assignment objects associated with this Batch
            that have not been completed.
        """
        return TaskAssignment.objects.filter(task__batch_id=self.id)\
                                     .filter(completed=False)

    def unfinished_tasks(self):
        """
        Returns:
            QuerySet of all Task objects associated with this Batch
            that have NOT been completed.
        """
        return self.task_set.filter(completed=False).order_by('id')

    def update_completed_status(self):
        """Update the `completed` flag to match the Batch completed status

        A Batch is marked "complete" IFF all of the Batch's Tasks are marked as `completed`.

        A database write occurs IFF the "completed" status has changed since the `completed`
        flag was last updated.
        """
        completed_status = not self.unfinished_tasks().exists()
        if self.completed != completed_status:
            self.completed = completed_status
            self.save()

    def users_that_completed_tasks(self):
        """
        Returns:
            QuerySet of all Users who have completed TaskAssignments
            that are part of this Batch
        """
        return User.objects.\
            filter(taskassignment__task__batch=self).\
            filter(taskassignment__completed=True).\
            distinct()

    def _parse_csv(self, csv_fh):
        """
        Args:
            csv_fh (file-like object): File handle for CSV output

        Returns:
            A tuple where the first value is a list of strings for the
            header fieldnames, and the second value is an iterable
            that returns a list of values for the rest of the row in
            the CSV file.
        """
        rows = csv.reader(csv_fh)
        header = next(rows)
        return header, rows

    def _get_csv_fieldnames(self, task_queryset):
        """
        Args:
            task_queryset (QuerySet):

        Returns:
            A tuple of strings specifying the fieldnames to be used in
            in the header of a CSV file.
        """
        input_field_set = set()
        answer_field_set = set()
        task_assignments = TaskAssignment.objects.\
            filter(task__in=task_queryset).\
            prefetch_related(Prefetch('task', queryset=task_queryset))
        for task_assignment in task_assignments:
            input_field_set.update(task_assignment.task.input_csv_fields.keys())

            # If the answers JSONField is empty, it evaluates as a string instead of a dict
            if task_assignment.answers != '':
                answer_field_set.update(task_assignment.answers.keys())
        return tuple(
            ['HITId', 'HITTypeId', 'Title', 'CreationTime', 'MaxAssignments',
             'AssignmentDurationInSeconds', 'AssignmentId', 'WorkerId',
             'AcceptTime', 'SubmitTime', 'WorkTimeInSeconds'] +
            ['Input.' + k for k in sorted(input_field_set)] +
            ['Answer.' + k for k in sorted(answer_field_set)] +
            ['Turkle.Username']
        )

    def _results_data(self, task_queryset):
        """
        All completed Tasks must come from the same project so that they have the
        same field names.

        Args:
            task_queryset (QuerySet):

        Returns:
            A tuple where the first value is a list of fieldname strings, and
            the second value is a list of dicts, where the keys to these
            dicts are the values of the fieldname strings.
        """
        rows = []
        time_format = '%a %b %d %H:%M:%S %Z %Y'
        task_assignments = TaskAssignment.objects.\
            filter(task__in=task_queryset).\
            filter(completed=True).\
            prefetch_related(Prefetch('task', queryset=task_queryset))
        for task_assignment in task_assignments:
            task = task_assignment.task
            batch = task.batch
            project = task.batch.project

            if task_assignment.assigned_to:
                username = task_assignment.assigned_to.username
            else:
                username = ''

            row = {
                'HITId': task.id,
                'HITTypeId': project.id,
                'Title': project.name,
                'CreationTime': batch.created_at.strftime(time_format),
                'MaxAssignments': batch.assignments_per_task,
                'AssignmentDurationInSeconds': batch.allotted_assignment_time * 3600,
                'AssignmentId': task_assignment.id,
                'WorkerId': task_assignment.assigned_to_id,
                'AcceptTime': task_assignment.created_at.strftime(time_format),
                'SubmitTime': task_assignment.updated_at.strftime(time_format),
                'WorkTimeInSeconds': task_assignment.work_time_in_seconds(),
                'Turkle.Username': username,
            }
            row.update({'Input.' + k: v for k, v in task.input_csv_fields.items()})
            row.update({'Answer.' + k: v for k, v in task_assignment.answers.items()})
            rows.append(row)

        return self._get_csv_fieldnames(task_queryset), rows

    def __str__(self):
        return 'Batch: {}'.format(self.name)


class TurklePermissionChecker(ObjectPermissionChecker):
    """
    Wrapper for django-guardian's permissions checker.

    Handles projects that don't have custom permissions.
    """

    def has_perm(self, perm, obj):
        """
        Checks if user/group has given permission for object.
        :param perm: permission as string, may or may not contain app_label
          prefix (if not prefixed, we grab app_label from ``obj``)
        :param obj: Django model instance for which permission should be checked
        """
        if self.user and not self.user.is_active:
            return False
        elif self.user and self.user.is_superuser:
            return True
        elif not self.user.is_authenticated and obj.login_required:
            return False
        elif obj.custom_permissions:
            if '.' in perm:
                _, perm = perm.split('.', 1)
            return perm in self.get_perms(obj)
        else:
            return True


class Project(TaskAssignmentStatistics, models.Model):
    class Meta:
        permissions = (
            # For consistency with Django Guardian naming conventions
            # ('{VERB}_{MODELNAME}'), this permission SHOULD have been
            # named 'can_work_on_project'
            ('can_work_on', 'Can work on Tasks for this Project'),
        )
        verbose_name = "Project"
        ordering = ['-id']

    active = models.BooleanField(db_index=True, default=True)
    allotted_assignment_time = models.IntegerField(default=24)
    assignments_per_task = models.IntegerField(db_index=True, default=1)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                   related_name='created_projects',
                                   on_delete=models.CASCADE, verbose_name='creator')
    custom_permissions = models.BooleanField(default=False)
    filename = models.CharField(max_length=1024, blank=True)
    html_template = models.TextField()
    html_template_has_submit_button = models.BooleanField(default=False)
    login_required = models.BooleanField(db_index=True, default=True)
    name = models.CharField(max_length=1024)
    updated_at = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True,
                                   related_name='updated_projects',
                                   on_delete=models.CASCADE)

    # Fieldnames are automatically extracted from html_template text
    fieldnames = JSONField(blank=True)

    def assignments_completed_by(self, user):
        """
        Returns:
            QuerySet of all TaskAssignments completed by specified user
            that are part of this Project
        """
        return TaskAssignment.objects.\
            filter(completed=True).\
            filter(assigned_to_id=user.id).\
            filter(task__batch__project=self)

    def available_for(self, user):
        """
        Returns:
            Boolean indicating if this Project is available for the user
        """
        if not user.is_authenticated and self.login_required:
            return False
        elif self.custom_permissions:
            return user.has_perm('can_work_on', self)
        else:
            return True

    def clean(self):
        super().clean()
        if len(self.html_template) > get_turkle_template_limit(True):
            raise ValidationError({'html_template': 'Template is too large'}, code='invalid')
        if not self.login_required and self.assignments_per_task != 1:
            raise ValidationError('When login is not required to access the Project, ' +
                                  'the number of Assignments per Task must be 1')
        self.process_template()

    def copy_permissions_to_batches(self):
        """Copy permissions from this Project to all associated Batches

        Copies:
        - `custom_permissions` flag
        - `login_required` flag
        - group-level access permissions
        """
        self.batch_set.update(
            custom_permissions=self.custom_permissions,
            login_required=self.login_required,
        )
        if self.custom_permissions:
            for group in get_groups_with_perms(self):
                if 'can_work_on' in get_group_perms(group, self):
                    batches = self.batch_set.all()
                    GroupObjectPermission.objects.bulk_assign_perm(
                        'can_work_on_batch', group, batches)

    def finished_task_assignments(self):
        """
        Returns:
            QuerySet of all Task Assignment objects associated with this Project
            that have been completed.
        """
        return TaskAssignment.objects.filter(task__batch__project_id=self.id)\
                                     .filter(completed=True)

    def process_template(self):
        soup = BeautifulSoup(self.html_template, 'html.parser')
        self.html_template_has_submit_button = bool(soup.select('input[type=submit]'))

        # Extract fieldnames from html_template text, save fieldnames as keys of JSON dict
        unique_fieldnames = set(re.findall(r'\${(\w+)}', self.html_template))
        self.fieldnames = dict((fn, True) for fn in unique_fieldnames)

        # Matching mTurk we confirm at least one input, select, or textarea
        if soup.find('input') is None and soup.find('select') is None \
                and soup.find('textarea') is None:
            msg = "Template does not contain any fields for responses. " + \
                  "Please include at least one field (input, select, or textarea)." + \
                  "This usually means you are generating HTML with JavaScript." + \
                  "If so, add an unused hidden input."
            raise ValidationError({'html_template': msg}, code='invalid')

    def total_assignments_completed_by(self, user):
        """
        Returns:
            Integer of total number of TaskAssignments completed by
            specified user that are part of this Project
        """
        return self.assignments_completed_by(user).count()

    def users_that_completed_tasks(self):
        """
        Returns:
            QuerySet of all Users who have completed TaskAssignments
            that are part of this Project
        """
        return User.objects.\
            filter(taskassignment__task__batch__project=self).\
            filter(taskassignment__completed=True).\
            distinct()

    def __str__(self):
        return self.name


class ActiveProjectManager(models.Manager):
    """Query projects by activity on assignments"""
    def get_queryset(self, **kwargs):
        n_days = int(kwargs.get('n_days', 7))
        time_cutoff = timezone.now() - timedelta(days=n_days)
        return super().get_queryset().\
            filter(Q(batch__task__taskassignment__updated_at__gt=time_cutoff) &
                   Q(batch__task__taskassignment__completed=True)).\
            distinct().\
            annotate(assignments=Count(
                'batch__task__taskassignment',
                filter=(Q(batch__task__taskassignment__updated_at__gt=time_cutoff) &
                        Q(batch__task__taskassignment__completed=True)))).\
            annotate(last_finished_time=Max(
                'batch__task__taskassignment__updated_at',
                filter=Q(batch__task__taskassignment__completed=True)))


class ActiveProject(Project):
    """Proxy object for Project using annotations from its manager"""
    objects = ActiveProjectManager()

    class Meta:
        proxy = True
        ordering = ['name']

    def completed_assignments(self):
        return self.assignments
    completed_assignments.admin_order_field = 'assignments'

    def most_recent(self):
        return self.last_finished_time
    most_recent.admin_order_field = 'last_finished_time'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-19 15:37
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Batch',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(db_index=True, default=True)),
                ('allotted_assignment_time', models.IntegerField(default=24)),
                ('assignments_per_task', models.IntegerField(default=1, verbose_name='Assignments per Task')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('filename', models.CharField(max_length=1024)),
                ('name', models.CharField(max_length=1024)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Batch',
                'verbose_name_plural': 'Batches',
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(db_index=True, default=True)),
                ('assignments_per_task', models.IntegerField(db_index=True, default=1)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('custom_permissions', models.BooleanField(default=False)),
                ('filename', models.CharField(blank=True, max_length=1024)),
                ('html_template', models.TextField()),
                ('html_template_has_submit_button', models.BooleanField(default=False)),
                ('login_required', models.BooleanField(db_index=True, default=True)),
                ('name', models.CharField(max_length=1024)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('fieldnames', jsonfield.fields.JSONField(blank=True)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='created_projects', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='updated_projects', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Project',
                'permissions': (('can_work_on', 'Can work on Tasks for this Project'),),
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('completed', models.BooleanField(default=False)),
                ('input_csv_fields', jsonfield.fields.JSONField()),
                ('batch', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='turkle.Batch')),
            ],
            options={
                'verbose_name': 'Task',
            },
        ),
        migrations.CreateModel(
            name='TaskAssignment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('answers', jsonfield.fields.JSONField(blank=True)),
                ('completed', models.BooleanField(db_index=True, default=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('expires_at', models.DateTimeField(null=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('assigned_to', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='turkle.Task')),
            ],
            options={
                'verbose_name': 'Task Assignment',
            },
        ),
        migrations.AddField(
            model_name='batch',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='turkle.Project'),
        ),
    ]



from datetime import datetime
import logging

from django.core.management.base import BaseCommand

from turkle.models import TaskAssignment


class Command(BaseCommand):
    help = ()

    def handle(self, *args, **options):
        t0 = datetime.now()
        (total_deleted, _) = TaskAssignment.expire_all_abandoned()
        t = datetime.now()
        dt = (t - t0).total_seconds()
        logging.basicConfig(format="%(asctime)-15s %(message)s", level=logging.INFO)
        logging.info('TURKLE: Expired {0} abandoned Task Assignments in {1:.3f} seconds'.
                     format(total_deleted, dt))



# Generated by Django 2.2.19 on 2021-04-07 15:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0009_batch_completed'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='project',
            options={'ordering': ['-id'], 'permissions': (('can_work_on', 'Can work on Tasks for this Project'),), 'verbose_name': 'Project'},
        ),
        migrations.AddField(
            model_name='project',
            name='allotted_assignment_time',
            field=models.IntegerField(default=24),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-09-23 15:30
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0003_batch_custom_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='batch',
            name='login_required',
            field=models.BooleanField(db_index=True, default=True),
        ),
    ]



from django.conf import settings

from . import __version__


def get_site_name():
    return getattr(settings, 'TURKLE_SITE_NAME', 'Turkle')


def get_turkle_template_limit(in_bytes=False):
    """get template size limit in kb by default"""
    template_size_limit = getattr(settings, 'TURKLE_TEMPLATE_LIMIT', 64)
    if in_bytes:
        template_size_limit *= 1024
    return template_size_limit


def turkle_vars(request):
    """add variables to the template context"""
    return {
        'turkle_site_name': get_site_name(),
        'turkle_template_limit': get_turkle_template_limit(),
        'turkle_version': __version__,
        'turkle_email_enabled': getattr(settings, 'TURKLE_EMAIL_ENABLED', False),
        'turkle_meta_tags': getattr(settings, 'META_TAGS', []),
    }


def are_anonymous_tasks_allowed():
    return getattr(settings, 'TURKLE_ANONYMOUS_TASKS', True)



# Generated by Django 2.2.20 on 2021-04-15 07:48

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management.sql import emit_post_migrate_signal
from django.db import migrations

from turkle.models import Batch, Project

User = get_user_model()


def add_turkle_user_admin_group(apps, schema_editor):
    """Create custom group that has CRU (but not D) access to User/Group models
    """

    # Force Django to create permissions before all migrations are run
    #
    # When running the migrations creates the database tables from
    # scratch (such as when running 'python manage.py test', but not
    # when running 'python manage.py migrate' to update an existing
    # database), Django will not add database rows with permissions
    # for the User and Group models until *after* all migrations have
    # been run.  We need these rows with permission data to be added
    # to the database now, so that we can grant the permissions to the
    # new Group instance we are creating.
    #
    # This is a known issue with Django that was identified in Django
    # bug #23422, "Cannot add Permission to Group in data migration"
    # back in 2014:
    #   https://code.djangoproject.com/ticket/23422
    #
    # The bug report was closed as "wontfix" in 2014.  The use of
    # emit_post_migrate_signal() - which supposedly works in Django
    # 2.2+ - was suggested in a comment on the bug report made on
    # 2020-05-07:
    #   https://code.djangoproject.com/ticket/23422#comment:28
    emit_post_migrate_signal(1, False, 'default')

    batch_ctype = ContentType.objects.get_for_model(Batch)
    group_ctype = ContentType.objects.get_for_model(Group)
    project_ctype = ContentType.objects.get_for_model(Project)
    user_ctype = ContentType.objects.get_for_model(User)

    batch_perms = [Permission.objects.get(codename='view_batch', content_type=batch_ctype)]
    group_perms = [Permission.objects.get(codename=codename, content_type=group_ctype)
                   for codename in ['add_group', 'change_group', 'view_group']]
    project_perms = [Permission.objects.get(codename='view_project', content_type=project_ctype)]
    user_perms = [Permission.objects.get(codename=codename, content_type=user_ctype)
                  for codename in ['add_user', 'change_user', 'view_user']]

    # Because we use get_or_create(), if an existing Turkle installation already
    # has a "Turkle User Admin" group, we will overwrite the existing permissions.
    (uag, _) = Group.objects.get_or_create(name='Turkle User Admin')
    uag.permissions.set(batch_perms + group_perms + project_perms + user_perms)


class Migration(migrations.Migration):

    dependencies = [
        ('turkle', '0010_project_allotted_assignment_time'),
    ]

    operations = [
        migrations.RunPython(add_turkle_user_admin_group),
    ]



from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def is_installed(app):
    return app in settings.INSTALLED_APPS


@register.filter
def meta_tag(tag):
    attributes = [f"{key}=\"{value}\"" for key, value in tag.items()]
    return mark_safe(f"<meta {' '.join(attributes)}>")


@register.filter()
def add_class(field, css):
    return field.as_widget(attrs={"class": css})



import setuptools

with open("turkle/__init__.py") as fp:
    ns = {}
    exec(fp.read(), ns)

with open("README.rst") as fp:
    long_description = fp.read()

with open('requirements.txt') as f:
    requirements = [line.strip() for line in f]

setuptools.setup(
    name="turkle",
    version=ns["__version__"],
    description="Django-based clone of Amazon's Mechanical Turk service",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/hltcoe/turkle",
    project_urls={
        "Documentation": "https://turkle.readthedocs.io/",
    },
    author="Craig Harman",
    author_email="craig@craigharman.net",
    license="BSD",
    packages=setuptools.find_packages(exclude=['scripts', 'turkle_site', 'turkle.tests']),
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    keywords=['annotation', 'crowdsourcing', 'mturk', 'mechanical turk'],
)
